from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict


@dataclass
class Message:
    """
    Dataclass abstraction for message
    """
    event: str
    payload: Dict[str, Any]
    ref: Any
    topic: str

    def __hash__(self):
        return hash((self.event, tuple(list(self.payload.values())), self.ref, self.topic))


class ChannelEvents(str, Enum):
    """
    ChannelEvents are a bunch of constant strings that are defined according to
    what the Phoenix realtime server expects.
    """

    close = "phx_close"
    error = "phx_error"
    join = "phx_join"
    reply = "phx_reply"
    leave = "phx_leave"
    heartbeat = "heartbeat"


PHOENIX_CHANNEL = "phoenix"
HEARTBEAT_PAYLOAD = {"msg": "ping"}



from typing import Callable, TypeVar

from typing_extensions import ParamSpec


T_ParamSpec = ParamSpec("T_ParamSpec")
T_Retval = TypeVar("T_Retval")

# Custom types
Callback = Callable[T_ParamSpec, T_Retval]



"""
Converts the change Payload into native Python types.
"""
import json
from dateutil.parser import parse


abstime = "abstime"
_bool = "bool"  # bool is a keyword in python
date = "date"
daterange = "daterange"
float4 = "float4"
float8 = "float8"
int2 = "int2"
int4 = "int4"
int4range = "int4range"
int8 = "int8"
int8range = "int8range"
_json = "json"  # Potentially a library
jsonb = "jsonb"
money = "money"
numeric = "numeric"
oid = "oid"
reltime = "reltime"
time = "time"
timestamp = "timestamp"
timestamptz = "timestamptz"
timetz = "timetz"
tsrange = "tsrange"
tstzrange = "tstzrange"


"""
Takes an array of columns and an object of string values then converts each string value
to its mapped type.

:param columns:
:param records:
:param options: The map of various options that can be applied to the mapper

"""


def convert_change_data(columns, records, options={}):
    skip_types = options.get("skip_types") if options.get(
        "skip_types") == "undefined" else []
    return {
        key: convert_column(key, columns, records, skip_types)
        for key in records.keys()
    }


def convert_column(column_name, columns, records, skip_types):
    column = next(filter(lambda x: x.get("name") == column_name, columns))
    if not column or column.get("type") in skip_types:
        return noop(records[column_name])
    else:
        return convert_cell(column.get("type"), records[column_name])


def convert_cell(_type: str, string_value: str):
    try:
        if string_value is None:
            return None
        # If data type is an array
        if _type[0] == "_":
            array_value = _type[1:len(_type)]
            return to_array(string_value, array_value)
        # If it's not null then we need to convert it to the correct type
        if _type == abstime:
            return noop(string_value)
        elif _type == _bool:
            print("converted to bool")
            return to_boolean(string_value)
        elif _type == date:
            return noop(string_value)
        elif _type == daterange:
            return to_date_range(string_value)
        elif _type == float4:
            return to_float(string_value)
        elif _type == float8:
            return to_float(string_value)
        elif _type == int2:
            return to_int(string_value)
        elif _type == int4:
            return to_int(string_value)
        elif _type == int4range:
            return to_int_range(string_value)
        elif _type == int8:
            return to_int(string_value)
        elif _type == int8range:
            return to_int_range(string_value)
        elif _type == _json:
            return to_json(string_value)
        elif _type == jsonb:
            return to_json(string_value)
        elif _type == money:
            return to_float(string_value)
        elif _type == numeric:
            return to_float(string_value)
        elif _type == oid:
            return to_int(string_value)
        elif _type == reltime:
            # To allow users to cast it based on Timezone
            return noop(string_value)
        elif _type == time:
            # To allow users to cast it based on Timezone
            return noop(string_value)
        elif _type == timestamp:
            return to_timestamp_string(
                string_value
            )  # Format to be consistent with PostgREST
        elif _type == timestamptz:
            return parse(string_value)
        elif _type == timetz:
            return parse(string_value)
        elif _type == tsrange:
            return to_date_range(string_value)
        elif _type == tstzrange:
            return to_date_range(string_value)
        else:
            # All the rest will be returned as strings
            return noop(string_value)

    except Exception as e:
        print(
            f"Could not convert cell of type {_type} and value {string_value}")
        print(f"This is the error {e}")
        return string_value


def noop(string_value: str):
    return string_value


def to_boolean(string_value: str):
    if string_value == "t":
        return True
    elif string_value == "f":
        return False
    else:
        return None


def to_date(string_value: str):
    return parse(string_value)


def to_date_range(string_value: str):
    arr = json.dumps(string_value)
    return [parse(arr[0]), parse(arr[1])]


def to_float(string_value):
    return float(string_value)


def to_int(string_value: str):
    return int(string_value)


def to_int_range(string_value: str):
    arr = json.loads(string_value)
    return [int(arr[0]), int(arr[1])]


def to_json(string_value: str):
    return json.loads(string_value)


"""
Converts a Postgres array into a native python list.
>>> to_array('{1,2,3,4}', 'int4')
    [1,2,3,4]
>>> to_array('{}', 'int4')
    []
"""


def to_array(string_value: str, type: str):
    # this takes off the '{' & '}'
    string_enriched = string_value[1: len(string_value) - 1]

    # Converts the string into an array
    # if string is empty (meaning the array was empty), an empty array will be immediately returned
    string_array = string_enriched.split(
        ",") if len(string_enriched) > 0 else []
    return list(map(lambda string: convert_cell(type, string), string_array))


"""
Fixes timestamp to be ISO-8601. Swaps the space between the date and time for a 'T'
>>> to_timestamp_string('2019-09-10 00:00:00')
    '2019-09-10T00:00:00'
"""


def to_timestamp_string(string_value: str):
    return string_value.replace(" ", "T")



import asyncio
import json
import logging
from collections import defaultdict
from functools import wraps
from typing import Any, Callable, List, Dict, TypeVar, DefaultDict

import websockets
from typing_extensions import ParamSpec

from realtime.channel import Channel
from realtime.exceptions import NotConnectedError
from realtime.message import HEARTBEAT_PAYLOAD, PHOENIX_CHANNEL, ChannelEvents, Message
from realtime.types import Callback, T_ParamSpec, T_Retval

logging.basicConfig(
    format="%(asctime)s:%(levelname)s - %(message)s", level=logging.INFO
)


def ensure_connection(func: Callback):
    @wraps(func)
    def wrapper(*args: T_ParamSpec.args, **kwargs: T_ParamSpec.kwargs) -> T_Retval:
        if not args[0].connected:
            raise NotConnectedError(func.__name__)

        return func(*args, **kwargs)

    return wrapper


class Socket:
    def __init__(
        self,
        url: str,
        auto_reconnect: bool = False,
        params: Dict[str, Any] = {},
        hb_interval: int = 5,
    ) -> None:
        """
        `Socket` is the abstraction for an actual socket connection that receives and 'reroutes' `Message` according to its `topic` and `event`.
        Socket-Channel has a 1-many relationship.
        Socket-Topic has a 1-many relationship.
        :param url: Websocket URL of the Realtime server. starts with `ws://` or `wss://`
        :param params: Optional parameters for connection.
        :param hb_interval: WS connection is kept alive by sending a heartbeat message. Optional, defaults to 5.
        """
        self.url = url
        self.channels = defaultdict(list)
        self.connected = False
        self.params = params
        self.hb_interval = hb_interval
        self.ws_connection: websockets.client.WebSocketClientProtocol
        self.kept_alive = False
        self.auto_reconnect = auto_reconnect

        self.channels: DefaultDict[str, List[Channel]] = defaultdict(list)

    @ensure_connection
    def listen(self) -> None:
        """
        Wrapper for async def _listen() to expose a non-async interface
        In most cases, this should be the last method executed as it starts an infinite listening loop.
        :return: None
        """
        loop = asyncio.get_event_loop()  # TODO: replace with get_running_loop
        loop.run_until_complete(asyncio.gather(self._listen(), self._keep_alive()))

    async def _listen(self) -> None:
        """
        An infinite loop that keeps listening.
        :return: None
        """
        while True:
            try:
                msg = await self.ws_connection.recv()
                msg = Message(**json.loads(msg))

                if msg.event == ChannelEvents.reply:
                    continue

                for channel in self.channels.get(msg.topic, []):
                    for cl in channel.listeners:
                        if cl.event in ["*", msg.event]:
                            cl.callback(msg.payload)
            except websockets.exceptions.ConnectionClosed:
                if self.auto_reconnect:
                    logging.info(
                        "Connection with server closed, trying to reconnect..."
                    )
                    await self._connect()
                    for topic, channels in self.channels.items():
                        for channel in channels:
                            await channel._join()
                else:
                    logging.exception("Connection with the server closed.")
                    break

    def connect(self) -> None:
        """
        Wrapper for async def _connect() to expose a non-async interface
        """
        loop = asyncio.get_event_loop()  # TODO: replace with get_running
        loop.run_until_complete(self._connect())
        self.connected = True

    async def _connect(self) -> None:
        ws_connection = await websockets.connect(self.url)

        if ws_connection.open:
            logging.info("Connection was successful")
            self.ws_connection = ws_connection
            self.connected = True
        else:
            raise Exception("Connection Failed")

    async def _keep_alive(self) -> None:
        """
        Sending heartbeat to server every 5 seconds
        Ping - pong messages to verify connection is alive
        """
        while True:
            try:
                data = dict(
                    topic=PHOENIX_CHANNEL,
                    event=ChannelEvents.heartbeat,
                    payload=HEARTBEAT_PAYLOAD,
                    ref=None,
                )
                await self.ws_connection.send(json.dumps(data))
                await asyncio.sleep(self.hb_interval)
            except websockets.exceptions.ConnectionClosed:
                if self.auto_reconnect:
                    logging.info(
                        "Connection with server closed, trying to reconnect..."
                    )
                    await self._connect()
                else:
                    logging.exception("Connection with the server closed.")
                    break

    @ensure_connection
    def set_channel(self, topic: str) -> Channel:
        """
        :param topic: Initializes a channel and creates a two-way association with the socket
        :return: Channel
        """
        chan = Channel(self, topic, self.params)
        self.channels[topic].append(chan)

        return chan

    def summary(self) -> None:
        """
        Prints a list of topics and event the socket is listening to
        :return: None
        """
        for topic, chans in self.channels.items():
            for chan in chans:
                print(f"Topic: {topic} | Events: {[e for e, _ in chan.callbacks]}]")



from __future__ import annotations

import asyncio
import json
from typing import Any, List, Dict, TYPE_CHECKING, NamedTuple

from realtime.types import Callback

if TYPE_CHECKING:
    from realtime.connection import Socket


class CallbackListener(NamedTuple):
    """A tuple with `event` and `callback` """
    event: str
    callback: Callback


class Channel:
    """
    `Channel` is an abstraction for a topic listener for an existing socket connection.
    Each Channel has its own topic and a list of event-callbacks that responds to messages.
    Should only be instantiated through `connection.Socket().set_channel(topic)`
    Topic-Channel has a 1-many relationship.
    """

    def __init__(self, socket: Socket, topic: str, params: Dict[str, Any] = {}) -> None:
        """
        :param socket: Socket object
        :param topic: Topic that it subscribes to on the realtime server
        :param params:
        """
        self.socket = socket
        self.params = params
        self.topic = topic
        self.listeners: List[CallbackListener] = []
        self.joined = False

    def join(self) -> Channel:
        """
        Wrapper for async def _join() to expose a non-async interface
        Essentially gets the only event loop and attempt joining a topic
        :return: Channel
        """
        loop = asyncio.get_event_loop()  # TODO: replace with get_running_loop
        loop.run_until_complete(self._join())
        return self

    async def _join(self) -> None:
        """
        Coroutine that attempts to join Phoenix Realtime server via a certain topic
        :return: None
        """
        join_req = dict(topic=self.topic, event="phx_join",
                        payload={}, ref=None)

        try:
            await self.socket.ws_connection.send(json.dumps(join_req))
        except Exception as e:
            print(str(e))  # TODO: better error propagation
            return

    def on(self, event: str, callback: Callback) -> Channel:
        """
        :param event: A specific event will have a specific callback
        :param callback: Callback that takes msg payload as its first argument
        :return: Channel
        """
        cl = CallbackListener(event=event, callback=callback)
        self.listeners.append(cl)
        return self

    def off(self, event: str) -> None:
        """
        :param event: Stop responding to a certain event
        :return: None
        """
        self.listeners = [
            callback for callback in self.listeners if callback.event != event]




__version__ = "1.0.1"

from realtime.channel import CallbackListener, Channel
from realtime.connection import Socket
from realtime.exceptions import NotConnectedError
from realtime.message import *
from realtime.transformers import *



class NotConnectedError(Exception):
    """
    Raised when operations requiring a connection are executed when socket is not connected
    """

    def __init__(self, func_name: str):
        self.offending_func_name: str = func_name

    def __str__(self):
        return f"A WS connection has not been established. Ensure you call Socket.connect() before calling Socket.{self.offending_func_name}()"
